{$I ALLLIB.DEF}

unit ZLIB112;

interface

uses
  SysUtils, Classes, ZLIBPas;

type

  TOnFileProgress = procedure(Sender: TObject; Msg: String; Position: Longint) of object;

  // TCustomZlibStream: Abstract ancestor class
  TCustomZlibStream = class(TStream)
  private
    FStrm: TStream;
    FStrmPos: Integer;
    FOnProgress: TOnFileProgress;
    FZRec: z_stream;
    FBuffer: array [Word] of Char;
  protected
    procedure Progress(Sender: TObject; Msg: String; Position: Longint); dynamic;
    property OnProgress: TOnFileProgress read FOnProgress write FOnProgress;
    constructor Create(Strm: TStream);
    property TheStream: TStream read FStrm;
  end;

  // TCompressionStream
  TCompressionStream = class(TCustomZlibStream)
  private
    function GetCompressionRate: Single;
  public
    constructor Create(Dest: TStream; Mode: Integer = -1 );
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    property CompressionRate: Single read GetCompressionRate;
    property OnProgress;
  end;

  // TDecompressionStream
  TDecompressionStream = class(TCustomZlibStream)
    FStreamOut: TStream;
    ResSize: Longint;
  public
    constructor Create(Source, Dest: TStream; AResSize: Longint);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    property OnProgress;
  end;

type
  EZlibError          = class(Exception);
  ECompressionError   = class(EZlibError);
  EDecompressionError = class(EZlibError);

var
  ZLIBUserAbort: Boolean = FALSE;

implementation

function CCheck(code: Integer): Integer;
begin
  Result := code;
  if code < 0 then
    raise ECompressionError.Create('error');
end;

function DCheck(code: Integer): Integer;
begin
  Result := code;
  if code < 0 then
    raise EDecompressionError.Create('error');
end;

// TCustomZlibStream -----------------------------------------------------------
constructor TCustomZLibStream.Create(Strm: TStream);
begin
  inherited Create;
  FStrm := Strm;
  FStrmPos := Strm.Position;
end;

procedure TCustomZLibStream.Progress(Sender: TObject; Msg: String; Position: Longint);
begin
  if Assigned(FOnProgress) then
    FOnProgress(Sender, Msg, Position);
end;


// TCompressionStream ----------------------------------------------------------
constructor TCompressionStream.Create(Dest: TStream; Mode: Integer = -1 );
begin
  inherited Create(Dest);
  FZRec.next_out := @FBuffer;
  FZRec.avail_out := sizeof(FBuffer);
  CCheck(deflateInit_(@FZRec, Mode, zlib_version, sizeof(FZRec)));
end;

destructor TCompressionStream.Destroy;
begin
  FZRec.next_in := nil;
  FZRec.avail_in := 0;
  try
    if FStrm.Position <> FStrmPos then FStrm.Position := FStrmPos;
    while not ZLIBUserAbort and (CCheck(deflate(FZRec, Z_FINISH)) <> Z_STREAM_END)
      and (FZRec.avail_out = 0) do
    begin
      FStrm.WriteBuffer(FBuffer, sizeof(FBuffer));
      FZRec.next_out := @FBuffer;
      FZRec.avail_out := sizeof(FBuffer);
      Progress(Self, 'write', sizeof(FBuffer)); // sizeof(FBuffer) bytes wrote
    end;
    if not ZLIBUserAbort and (FZRec.avail_out < sizeof(FBuffer)) then
    begin
      FStrm.WriteBuffer(FBuffer, sizeof(FBuffer) - FZRec.avail_out);
      Progress(Self, 'write', sizeof(FBuffer) - FZRec.avail_out); // dif. bytes wrote
    end;
  finally
    deflateEnd(FZRec);
  end;
  inherited Destroy;
end;

function TCompressionStream.Read(var Buffer; Count: Longint): Longint;
begin
  raise ECompressionError.Create('Invalid stream operation');
end;

function TCompressionStream.Write(const Buffer; Count: Longint): Longint;
begin
  Progress(Self, 'read', Count);// count bytes were read
  FZRec.next_in := @Buffer;
  FZRec.avail_in := Count;
  if FStrm.Position <> FStrmPos then FStrm.Position := FStrmPos;
  while not ZLIBUserAbort and (FZRec.avail_in > 0) do
  begin
    CCheck(deflate(FZRec, 0));
    //Progress(Self, 'compressed', ?); // compressed n bytes
    if FZRec.avail_out = 0 then
    begin
      FStrm.WriteBuffer(FBuffer, sizeof(FBuffer));
      FZRec.next_out := @FBuffer;
      FZRec.avail_out := sizeof(FBuffer);
      FStrmPos := FStrm.Position;
      Progress(Self, 'write', sizeof(FBuffer));// sizeof(FBuffer) bytes wrote
    end;
  end;
  Result := Count;
end;

function TCompressionStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  if (Offset = 0) and (Origin = soFromCurrent) then
    Result := FZRec.total_in
  else
    raise ECompressionError.Create('Invalid stream operation');
end;

function TCompressionStream.GetCompressionRate: Single;
begin
  if FZRec.total_in = 0 then
    Result := 0
  else
    Result := (1.0 - (FZRec.total_out / FZRec.total_in)) * 100.0;
end;


// TDecompressionStream --------------------------------------------------------
constructor TDecompressionStream.Create(Source, Dest: TStream; AResSize: Longint);
begin
  inherited Create(Source);
  FZRec.next_in := @FBuffer;
  FZRec.avail_in := 0;
  ResSize := AResSize;
  FStreamOut := Dest;
  DCheck(inflateInit_(@FZRec, zlib_version, sizeof(FZRec)));
end;

destructor TDecompressionStream.Destroy;
begin
  inflateEnd(FZRec);
  inherited Destroy;
end;

function TDecompressionStream.Read(var Buffer; Count: Longint): Longint;

  procedure Put(NoBytes: Longint);
  begin
    if Assigned(FStreamOut) then
    begin
      FStreamOut.Write(Buffer, NoBytes);
      Progress(Self, 'write', NoBytes); // NoBytes bytes wrote
    end;
  end;

begin
  FZRec.next_out := @Buffer;
  FZRec.avail_out := Count;
  if FStrm.Position <> FStrmPos then
    FStrm.Position := FStrmPos;
  while not ZLIBUserAbort and (FZRec.avail_out > 0) do
  begin
    if FZRec.avail_in = 0 then
    begin

      if Count > SizeOf(FBuffer) then // LR! May 17, 2001
        Count := SizeOf(FBuffer);     // LR! May 17, 2001
        
      if ResSize > 0 then
	if ResSize > Count then
	  FZRec.avail_in := FStrm.Read(FBuffer, Count)
	else
	  FZRec.avail_in := FStrm.Read(FBuffer, ResSize);

      if FZRec.avail_in > ResSize then
	FZRec.avail_in := ResSize;
      Dec(ResSize, FZRec.avail_in);
      Progress(Self, 'read', FZRec.avail_in); // avail_in bytes were read (or more, but this is what we'll process)

      if FZRec.avail_in = 0 then
        begin
          Result := Count - FZRec.avail_out;
          Put(Result);
          Exit;
        end;
      FZRec.next_in := @FBuffer;
      FStrmPos := FStrm.Position;
    end;
    try
      if not ZLIBUserAbort then
      begin
        CCheck(inflate(FZRec, 0));
        //Progress(Self, 'decompressed', Count-FZRec.avail_out); // decompressed
      end;
    except
      on E: Exception do ;
    end;
  end;
  Result := Count;
  Put(Result);
end;

function TDecompressionStream.Write(const Buffer; Count: Longint): Longint;
begin
  raise EDecompressionError.Create('Invalid stream operation');
end;

function TDecompressionStream.Seek(Offset: Longint; Origin: Word): Longint;
var
  I: Integer;
  Buf: array [0..4095] of Char;
begin
  if (Offset = 0) and (Origin = soFromBeginning) then
  begin
    DCheck(inflateReset(FZRec));
    FZRec.next_in := @FBuffer;
    FZRec.avail_in := 0;
    FStrm.Position := 0;
    FStrmPos := 0;
  end
  else if ( (Offset >= 0) and (Origin = soFromCurrent)) or
          ( ((Offset - FZRec.total_out) > 0) and (Origin = soFromBeginning)) then
  begin
    if Origin = soFromBeginning then Dec(Offset, FZRec.total_out);
    if Offset > 0 then
    begin
      for I := 1 to Offset div sizeof(Buf) do
	     ReadBuffer(Buf, sizeof(Buf));
      ReadBuffer(Buf, Offset mod sizeof(Buf));
    end;
  end
  else
    raise EDecompressionError.Create('Invalid stream operation');
  Result := FZRec.total_out;
end;

end.

