unit ZLibComp;

{$WARNINGS OFF}

{*********************************************************}
{* ZLib compression. This unit is provided as freeware   *}
{* and is written for 32-bit Delphi (D2 and higher).     *}
{*                                                       *}
{* As noted in the source, portions are copyrighted by   *}
{* the following parties:                                *}
{*   Copyright (c) 1997 Borland International            *}
{*   Copyright (c) 2000-2001 David O. Martin             *}
{*********************************************************}

{-------------------------------------------------------------}
{ Version history:                                            }
{   09/19/00  Version 1.0                                     }
{   10/30/00  Version 1.1. Moved CompressBuf and              }
{             DecompressBuf to the interface section.         }
{   03/08/01  Version 1.2. Added ZLibCompressStringFxn,       }
{             ZLibDecompressStringFxn, ZLibCompressMemStream, }
{             ZLibDecompressMemStream.                        }
{-------------------------------------------------------------}

interface

uses Classes;

const
  Version = '1.2';

procedure ZLibCompressString(var s: string; compressionLevel: Integer);
{ Compress string s using zlib (thread-safe). The compressionLevel may vary
  from 1 (best speed but worst compression) to 9 (slowest speed but best
  compression). A compressionLevel of 1 will usually give good speed with good
  compression. However, a higher compression level may yield better
  performance when bandwith is at a premium because the total data transfer
  time is given by the sum of compression time + download time + decompression
  time.

  NOTE: Errors may be trapped by placing this routine between a
  try..except block. (This ***should definitely*** be done on a
  server to prevent an exception dialog from halting the server.) }

procedure ZLibDecompressString(var s: string);
{ Decompress string s that was compressed using ZLibCompressString.

  NOTE: Errors may be trapped by placing this routine between a
  try..except block. (This ***should definitely*** be done on a
  server to prevent an exception dialog from halting the server.) }

function ZLibCompressStringFxn(const s: string; compressionLevel: Integer): string;
{ Like ZLibCompressString, but compressed string is returned as the
  function result. If it is not necessary to save the uncompressed string,
  then one should call ZLibCompressString as it is more memory efficient. }

function ZLibDecompressStringFxn(const s: string): string;
{ Like ZLibDeompressString, but decompressed string is returned as the
  function result. If it is not necessary to save the compressed string,
  then one should call ZLibDecompressString as it is more memory efficient. }

procedure ZLibCompressMemStream(s: TMemoryStream; compressionLevel: integer);
{ Compresses a memory stream. compressionLevel is as defined in
  ZLibCompressString. }

procedure ZLibDecompressMemStream(s: TMemoryStream);
{ Decompresses a memory stream. }

{------------------------------------------------------------------}
{ Driver routines for the above (which can be useful on there own) }
{------------------------------------------------------------------}

procedure CompressBuf(
  const inBuf: Pointer;     { input: pointer to source data }
        inBytes: Integer;   { input: # of bytes in inBuf }
        level: Integer;     { input: compression level }
    var outBuf: Pointer;    { output: pointer to newly allocated buffer with compressed data }
    var outBytes: Integer); { output: # of bytes in outBuf }
{ This routine compresses data. Level may vary from 0 to 9 (3 to 6 may
  be a reasonable trade-off between speed and compression).

  Compression levels:
    0 = no compression                 (Z_NO_COMPRESSION constant)
    1 = best speed                     (Z_BEST_SPEED constant)
    9 = best compression, but slowest  (Z_BEST_COMPRESSION constant) }

procedure DecompressBuf(
  const inBuf: Pointer;     { input: pointer to compressed data }
        inBytes: Integer;   { input: # of bytes in inBuf }
    var outBuf: Pointer;    { output: pointer to newly allocated buffer with decompressed data }
    var outBytes: Integer); { output: # of bytes in outBuf }
{ This routine decompresses data that was compressed with CompressBuf. Note
  that this works for all compression levels. }

implementation

uses Windows, SysUtils, ZLibPas;

{----------------------------------------------------------}
{ The CompressBuf and DecompressBuf routines that were     }
{ shipped with Delphi 3-5 are based on zlib 1.0.4 and are  }
{ not compatible with zlib 1.1.2. The following are        }
{ rewritten versions of CompressBuf and DecompressBuf that }
{ take into account the differences between zlib 1.0.4 and }
{ 1.1.2. Testing has shown this version is significantly   }
{ faster than the zlib code distributed with Delphi 5.     }
{                                                          }
{ Copyright (c) 1997 Borland International                 }
{ Copyright (c) 2000-2001 David O. Martin                  }
{----------------------------------------------------------}

procedure CompressBuf(
  const inBuf: Pointer;     { input: pointer to source data }
        inBytes: Integer;   { input: # of bytes in inBuf }
        level: Integer;     { input: compression level }
    var outBuf: Pointer;    { output: pointer to newly allocated buffer with compressed data }
    var outBytes: Integer); { output: # of bytes in outBuf }
{ This routine compresses data. Level may vary from 0 to 9 (3 to 6 may
  be a reasonable trade-off between speed and compression).

  Compression levels:
    0 = no compression                 (Z_NO_COMPRESSION constant)
    1 = best speed                     (Z_BEST_SPEED constant)
    9 = best compression, but slowest  (Z_BEST_COMPRESSION constant) }

  function Check(code: Integer): Integer;
  begin
    Result := code;
    if (code < 0) then
      raise Exception.Create('Compression error: '+z_errmsg[2-code]);
  end;

var
  strm: z_stream;
begin
  FillChar(strm, sizeof(strm), 0);
  outBytes := ((inBytes + (inBytes div 10) + 12) + 255) and not 255;
  GetMem(outBuf, outBytes);
  try
    strm.next_in := inBuf;
    strm.avail_in := inBytes;
    strm.next_out := outBuf;
    strm.avail_out := outBytes;
    Check(deflateInit(strm, level));
    try
      while Check(deflate(strm, Z_FINISH)) <> Z_STREAM_END do begin
        Inc(outBytes, 256);
        ReallocMem(outBuf, outBytes);
        strm.next_out := pBytef(Integer(outBuf) + strm.total_out);
        strm.avail_out := 256;
      end;
    finally
      Check(deflateEnd(strm));
    end;
    ReallocMem(outBuf, strm.total_out);
    outBytes := strm.total_out;
  except
    DeAllocMem(outBuf);
    raise;
  end;
end;

procedure DecompressBuf(
  const inBuf: Pointer;     { input: pointer to compressed data }
        inBytes: Integer;   { input: # of bytes in inBuf }
    var outBuf: Pointer;    { output: pointer to newly allocated buffer with decompressed data }
    var outBytes: Integer); { output: # of bytes in outBuf }
{ This routine decompresses data that was compressed with CompressBuf. }

  function Check(code: Integer): Integer;
  begin
    Result := code;
    if (code < 0) then
      raise Exception.Create('Decompression error: '+z_errmsg[2-code]);
  end;

var
  strm: z_stream;
  bufInc: Integer;
begin
  FillChar(strm, sizeof(strm), 0);
  bufInc := (inBytes + 255) and not 255;
  outBytes := bufInc;
  GetMem(outBuf, outBytes);
  try
    strm.next_in := inBuf;
    strm.avail_in := inBytes;
    strm.next_out := outBuf;
    strm.avail_out := outBytes;
    Check(inflateInit(strm));
    try
      while Check(inflate(strm, Z_NO_FLUSH)) <> Z_STREAM_END do begin
        Inc(outBytes, bufInc);
        ReallocMem(outBuf, outBytes);
        strm.next_out := pBytef(Integer(outBuf) + strm.total_out);
        strm.avail_out := bufInc;
      end;
    finally
      Check(inflateEnd(strm));
    end;
    ReallocMem(outBuf, strm.total_out);
    outBytes := strm.total_out;
  except
    DeAllocMem(outBuf);
    raise;
  end;
end;

procedure ZLibCompressString{(See interface)};
var
  outBuf: Pointer;
  outBytes: Integer;
begin
  CompressBuf(@s[1], Length(s), compressionLevel, outBuf, outBytes);
  SetLength(s, outBytes);
  Windows.CopyMemory(@s[1], outBuf, outBytes);
  DeAllocMem(outBuf);
end;

procedure ZLibDecompressString{(See interface)};
var
  outBuf: Pointer;
  outBytes: Integer;
begin
  DecompressBuf(@s[1], Length(s), outBuf, outBytes);
  SetLength(s, outBytes);
  Windows.CopyMemory(@s[1], outBuf, outBytes);
  DeAllocMem(outBuf);
end;

function ZLibCompressStringFxn{(See interface)};
begin
   Result:=s;
   ZLibCompressString(Result, compressionLevel);
end;

function ZLibDecompressStringFxn{(See interface)};
begin
   Result:=s;
   ZLibDecompressString(Result);
end;

procedure ZLibCompressMemStream{(See interface)};
var
  inBuf, outBuf: Pointer;
  inBytes, outBytes: Integer;
begin
  inBuf:=nil;
  outBuf:=nil;
  try
    inBytes:=s.Size;
    GetMem(inBuf, inBytes);
    s.Position:=0;
    s.ReadBuffer(inBuf^, inBytes);
    s.Clear;
    CompressBuf(inBuf, inBytes, compressionLevel, outBuf, outBytes);
    s.WriteBuffer(outBuf^, outBytes);
  finally
    DeAllocMem(inBuf);
    DeAllocMem(outBuf);
  end;
end;

procedure ZLibDecompressMemStream{(See interface)};
var
  inBuf, outBuf: Pointer;
  inBytes, outBytes: Integer;
begin
  inBuf:=nil;
  outBuf:=nil;
  try
    inBytes:=s.Size;
    GetMem(inBuf, inBytes);
    s.Position:=0;
    s.ReadBuffer(inBuf^, inBytes);
    s.Clear;
    DecompressBuf(inBuf, inBytes, outBuf, outBytes);
    s.WriteBuffer(outBuf^, outBytes);
  finally
    DeAllocMem(inBuf);
    DeAllocMem(outBuf);
  end;
end;

end.
